import json
import os

from cnvrgv2.data.remote_files_handler import RemoteFilesHandler
from cnvrgv2.proxy import HTTP
from cnvrgv2.utils.storage_utils import create_dir_if_not_exists
from cnvrgv2.utils.url_utils import urljoin


class FileDownloader(RemoteFilesHandler):
    def __init__(self, data_owner, use_cached=False, num_workers=40, queue_size=5000, chunk_size=1000,
                 override=False, progress_bar_enabled=False, query_slug=None, fullpath_filter=None):
        """
        Multithreaded file bucket_downloader.py - download files from server
        @param data_owner: Cnvrg dataset / project object
        @param num_workers: Number of threads to handle files
        @param queue_size: Max number of file meta to put in queue
        @param chunk_size: File meta chunk size to fetch from the server
        @param override: Override existing files
        """
        self.mode = "files"
        self.commits_url = urljoin(data_owner._route, "commits", data_owner.last_commit,
                                   "files") + "?" + "page[after]={}&page[size]={}&sort=id&mode={}"

        self.query_slug = query_slug

        # Default filter matches all files
        self.filter = {
            "operator": 'OR',
            "conditions": [
                {
                    "key": 'fullpath',
                    "operator": 'like',
                    "value": "*",
                }
            ],
        }

        total_files_query_url = self.commits_url.format('', 0, "files")
        total_folders_query_url = self.commits_url.format('', 0, "folders")

        if fullpath_filter:
            self.filter["conditions"][0]["value"] = fullpath_filter
            total_files_query_url += "&filter={}".format(json.dumps(self.filter))
            total_folders_query_url += "&filter={}".format(json.dumps(self.filter))

        if query_slug:
            total_files_query_url += "&query_slug={}".format(query_slug)
            total_folders_query_url += "&query_slug={}".format(query_slug)

        super().__init__(
            data_owner,
            num_workers=num_workers,
            queue_size=queue_size,
            chunk_size=chunk_size,
            override=override,
            use_cached=use_cached,
            progress_bar_enabled=progress_bar_enabled
        )

        total_files = data_owner._proxy.call_api(
            route=total_files_query_url,
            http_method=HTTP.GET
        ).meta["total"]

        total_folders = data_owner._proxy.call_api(
            route=total_folders_query_url,
            http_method=HTTP.GET
        ).meta["total"]

        self.total_files = total_files + total_folders

    def _collector_function(self, page_after=None):
        """
        Function to collect files that should be downloaded
        @param page_after: The id of the next file that the iteration of the pagination should start from
        @return: Should return array of files metadata
        """

        data = {
            "filter": json.dumps(self.filter),
            "query_slug": self.query_slug,
            "cache_link": self.use_cached
        }

        response = self.data_owner._proxy.call_api(
            route=self.commits_url.format(page_after, self.chunk_size, self.mode),
            http_method=HTTP.GET,
            payload=data
        )

        # Once we are out of files we want to move to downloading (empty) folders
        if response.meta["next"] == '' and self.mode == "files":
            self.mode = "folders"
            return {
                "file_dict": [],
                "total_files": self.total_files,
                "total_files_size": 0,
                "next": -100  # placeholder in order to nullify the cursor
            }

        file_dict = []
        for file in response.items:
            file_dict.append(dict(file.attributes))

        return {
            "file_dict": file_dict,
            "total_files": self.total_files,
            "next": response.meta["next"],
            "total_files_size": response.meta["total_files_size"]
        }

    def _handle_file_function(self, local_path, progress_bar=None, **kwargs):
        """
        Function that download single file
        @param local_path: File location locally
        @param progress_bar: A progress bar object to be used during the download
        @param kwargs: Needs to be object_path of the file in the bucket
        @return: None
        """
        is_cached = False
        if self.use_cached:
            is_cached = cache_file(self.data_owner.title, local_path, kwargs["fullpath"],
                                   kwargs.get("cached_commits", []))

        if not is_cached and (self.override or not os.path.exists(local_path)):
            self.storage_client.download_single_file(local_path, kwargs["object_path"], progress_bar=progress_bar)

        self.handle_queue.task_done()
        self.progress_queue.put(local_path)


def cache_file(dataset, local_path, fullpath, commits, back_path="/nfs-disk/"):
    is_cached = False
    for commit in commits:
        nfs_path = "{}{}/{}/{}".format(back_path, commit, dataset, fullpath)
        if os.path.exists(nfs_path):
            is_cached = True
            create_dir_if_not_exists(local_path)
            os.link(os.path.abspath(nfs_path), local_path)

    return is_cached



import click

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.cli.utils.helpers import build_grid_url, callback_log, parse_parameters_from_file
from cnvrgv2.cli.utils.helpers import print_generator_in_chunks
from cnvrgv2.cli.utils.options import PythonLiteralOption
from cnvrgv2.modules.workflows.workflow_utils import WorkflowStatuses, WorkflowUtils


@click.group(name='experiment')
def experiment_group():
    pass


@experiment_group.command()
@click.option('-t', '--title', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-tm', '--templates', default=None, help=messages.EXPERIMENT_HELP_TEMPLATES)
@click.option('-l/-nl', '--local/--no-local', default=False, help=messages.EXPERIMENT_HELP_LOCAL)
@click.option('-c', '--command', prompt=messages.EXPERIMENT_PROMPT_COMMAND, help=messages.EXPERIMENT_HELP_COMMAND)
@click.option('-d', '--datasets', cls=PythonLiteralOption, default=[], help=messages.EXPERIMENT_HELP_DATASETS)
@click.option('-ds', '--datasources', cls=PythonLiteralOption, default=[], help=messages.EXPERIMENT_HELP_DATASOURCES)
@click.option('-v', '--volume', default=None, help=messages.EXPERIMENT_HELP_VOLUME)
@click.option('-sb/-nsb', '--sync-before/--no-sync-before', default=True, help=messages.EXPERIMENT_HELP_SYNC_BEFORE)
@click.option('-sa/-nsa', '--sync-after/--no-sync-after', default=True, help=messages.EXPERIMENT_HELP_SYNC_AFTER)
@click.option('-i', '--image', default=None, help=messages.EXPERIMENT_HELP_IMAGE)
@click.option('-gb', '--git-branch', default=None, help=messages.EXPERIMENT_HELP_GIT_BRANCH)
@click.option('-gc', '--git-commit', default=None, help=messages.EXPERIMENT_HELP_GIT_COMMIT)
@click.option('-od', '--output_dir', default=None, help=messages.EXPERIMENT_HELP_OUTPUT_DIR)
@click.option('-g', '--grid', default=None, help=messages.EXPERIMENT_HELP_GRID)
@click.option('-lf', '--local_folders', default=None, help=messages.EXPERIMENT_HELP_LOCAL_FOLDERS)
@click.option('-log', '--log', default=False, help=messages.EXPERIMENT_HELP_LOG)
@click.option('-w', '--wait', default=False, help=messages.EXPERIMENT_HELP_LOG)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def run(
        cnvrg,
        logger,
        project,
        title,
        templates,
        local,
        command,
        datasets,
        volume,
        sync_before,
        sync_after,
        output_dir,
        image,
        git_branch,
        git_commit,
        grid,
        local_folders,
        log,
        wait,
        project_slug,
        datasources
):
    """
      run an experiment
    """
    dataset_objects = []
    volume_object = None
    kwargs = {}
    templates_list = templates.split(",") if templates else None

    for dataset in datasets:
        if dataset.get("commit") and dataset.get("query"):
            logger.log_and_echo(messages.EXPERIMENT_DATASET_QUERY_ERROR_MESSAGE.format(dataset.get("slug")),
                                error=True)
            return
        ds = cnvrg.datasets.get(dataset.get("slug"))
        if dataset.get("commit"):
            ds.local_commit = dataset.get("commit")
        if dataset.get("query"):
            ds.query = dataset.get("query")
        dataset_objects.append(ds)

    if volume:
        volume_object = project.volumes.get(volume)

    if image:
        image_name, image_tag = image.split(":")
        kwargs["image"] = cnvrg.images.get(name=image_name, tag=image_tag)

    if git_branch:
        kwargs["git_branch"] = git_branch

    if git_commit:
        kwargs["git_commit"] = git_commit

    if output_dir:
        kwargs["output_dir"] = output_dir

    if local_folders:
        kwargs["local_folders"] = local_folders

    if grid:
        parsed_parameters = parse_parameters_from_file(grid)
        kwargs["parameters"] = parsed_parameters
        grid_slug = project.experiments.create_grid(
            title=title,
            templates=templates_list,
            command=command,
            datasets=dataset_objects,
            volume=volume_object,
            sync_before=sync_before,
            **kwargs
        )
        grid_url = build_grid_url(cnvrg, project, grid_slug)
        success_message = messages.EXPERIMENT_GRID_CREATE_SUCCESS.format(grid_slug, grid_url)
        logger.log_and_echo(success_message)
    else:
        experiment = project.experiments.create(
            title=title,
            templates=templates_list,
            local=local,
            command=command,
            datasets=dataset_objects,
            volume=volume_object,
            sync_before=sync_before,
            sync_after=sync_after,
            **kwargs
        )
        success_message = messages.EXPERIMENT_CREATE_SUCCESS.format(experiment.title, experiment.full_href)
        logger.log_and_echo(success_message)

        if log or wait:
            WorkflowUtils.wait_for_statuses(experiment, WorkflowStatuses.FINAL_STATES,
                                            callback=callback_log(experiment, log))


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-c', '--commit', prompt=messages.EXPERIMENT_PROMPT_COMMIT, default='latest', required=False,
              help=messages.EXPERIMENT_HELP_COMMIT)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def merge_to_master(logger, project, experiment, slug, commit, project_slug):
    # TODO: Slug is not necessary as a parameter (it is as an option, since prepare command uses it).
    #  Change prepare command to send only required arguments

    if project.git:
        logger.log_and_echo(messages.EXPERIMENT_GIT_ERROR_MESSAGE, error=True)
        return
    commit_param = None if commit == 'latest' else commit
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    sha1 = experiment.merge_to_master(commit_param)
    logger.log_and_echo(messages.EXPERIMENT_MERGE_SUCCESS)
    logger.log_and_echo(messages.COMMIT_SHA1_MESSAGE.format(sha1))


@experiment_group.command()
@click.option('-k', '--key', help='')
@click.option('-v', '--value', help='')
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def log_param(experiment, logger, slug, key, value, project_slug):
    """
      logging a parameter of an experiment
    """
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    experiment.log_param(key=key, value=value)
    logger.log_and_echo('{0}: {1} was logged'.format(key, value))


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-f', '--files', prompt=messages.DATASET_PUT_PROMPT_FILES, help=messages.DATASET_PUT_HELP_FILES)
@click.option('-g', '--git-diff', is_flag=True, default=None, help=messages.DATA_UPLOAD_HELP_GIT_DIFF)
@click.option('-d', '--work-dir', required=False, help=messages.EXPERIMENT_HELP_COMMIT)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def log_artifacts(cnvrg, logger, experiment, slug, files, git_diff, work_dir, project_slug):
    file_paths = files.split(",")
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    experiment.log_artifacts(paths=file_paths, git_diff=git_diff, work_dir=work_dir)
    logger.log_and_echo(messages.EXPERIMENT_LOG_ARTIFACTS_SUCCESS)


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-f', '--files', prompt=messages.DATASET_PUT_PROMPT_FILES, help=messages.DATASET_PUT_HELP_FILES)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def log_images(cnvrg, logger, experiment, slug, files, project_slug):
    file_paths = files.split(",")
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    experiment.log_images(file_paths=file_paths)
    logger.log_and_echo(messages.EXPERIMENT_LOG_IMAGES_SUCCESS)


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def start_tensorboard(experiment, logger, slug, project_slug):
    """
      start a tensorboard session for current experiment
    """
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    experiment.start_tensorboard()
    logger.log_and_echo('tensorboard started!')


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def stop_tensorboard(experiment, logger, slug, project_slug):
    """
      stop a tensorboard session for current experiment
    """
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    experiment.stop_tensorboard()
    logger.log_and_echo('tensorboard session stopped')


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-c', '--commit', prompt=messages.EXPERIMENT_ARTIFACTS_PROMPT_COMMIT, default='latest', required=False,
              help=messages.EXPERIMENT_ARTIFACTS_HELP_COMMIT)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def pull_artifacts(experiment, logger, slug, commit, project_slug):
    """
      pull artifacts
      default is experiment's last commit
    """
    if experiment is None:
        logger.log_and_echo(messages.EXPERIMENT_DOES_NOT_EXIST, error=True)
    commit_param = None if commit == 'latest' else commit
    experiment.pull_artifacts(commit_sha1=commit_param)
    logger.log_and_echo(messages.EXPERIMENT_PULL_ARTIFACTS_SUCCESS)


@experiment_group.command()
@click.option('-s', '--slug', default=None, help=messages.EXPERIMENT_HELP_TITLE)
@click.option('-d', '--delete_artifacts', default=False, help=messages.EXPERIMENT_DELETE_ARTIFACTS, required=False)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def delete(experiment, logger, slug, delete_artifacts, project_slug):
    """
      delete experiment by slug
    """
    if delete_artifacts:
        user_answer = click.confirm(messages.EXPERIMENT_DELETE_ARTIFACTS_PROMPT, abort=False)
        if not user_answer:
            logger.log_and_echo(messages.EXPERIMENT_DELETE_ARTIFACTS_ABORTED, error=True)
            return

    experiment.delete(delete_artifacts=delete_artifacts)
    logger.log_and_echo(messages.EXPERIMENT_DELETE_SUCCESS)


@experiment_group.command()
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@click.option('-cs', '--chunk-size', default=20, type=int, help='Number of experiments to display at a time')
@click.option('-l', '--limit', type=int, help='Maximum number of experiments to display')
@prepare_command()
def list(cnvrg, project_slug, chunk_size, limit):
    project = cnvrg.projects.get(project_slug)
    experiments = project.experiments.list()
    attributes_field_name = [
        "Experiment Name",
        "Created By",
        "Experiment Slug",
        "Experiment Command",
        "Image Used",
        "Experiment Status",
        "Compute"
    ]
    attributes = ["title", "username", "slug", "input", "image_name", "status", "compute_name"]
    print_generator_in_chunks(experiments, chunk_size, limit, attributes, attributes_field_name, line_numbers=True)



import pytest

from cnvrgv2.proxy import Proxy
from cnvrgv2.errors import CnvrgError
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.modules.registries import Registry, RegistryTypes


class MockRegistry:
    def __init__(self):
        self.slug = "test-slug"
        self.url = "doker.io"
        self.type = RegistryTypes.OTHER
        self.title = "my-registry"
        self.username = "username"

        self.data = {
            "id": "17",
            "type": "registry",
            "attributes": {
                "slug": self.slug,
                "url": self.url,
                "title": self.title,
                "private": True,
                "username": self.username,
                "registry_type": self.type
            }
        }

    def response(self, *args, **kwargs):
        return JAF(response={"data": self.data})


class TestImage:
    def test_registry_attributes(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockRegistry()
        mocked_call = mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)

        registry = Registry(empty_cnvrg, mock_resp.slug)
        assert registry.slug == mock_resp.slug
        assert registry.url == mock_resp.url
        assert registry.title == mock_resp.title
        assert registry.registry_type == mock_resp.type

        assert mocked_call.call_count == 1
        assert mocked_call.call_args[1]['http_method'] == 'GET'

    def test_delete_registry(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockRegistry()
        mocked_call = mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)

        registry = Registry(empty_cnvrg, mock_resp.slug)
        registry.delete()

        assert mocked_call.call_count == 1
        assert mocked_call.call_args[1]['http_method'] == 'DELETE'

    def test_update_registry(self, mocker, empty_context, empty_cnvrg):
        mock_resp = MockRegistry()
        mocked_call = mocker.patch.object(Proxy, 'call_api', side_effect=mock_resp.response)

        registry = Registry(empty_cnvrg, mock_resp.slug)
        registry.update(url="new.url")

        assert mocked_call.call_count == 1
        assert mocked_call.call_args[1]['http_method'] == 'PUT'

    def test_update_registry_invalid_arguments(self, empty_context, empty_cnvrg):
        registry = Registry(empty_cnvrg, "slug")

        with pytest.raises(CnvrgError):
            registry.update(url=1)

        with pytest.raises(CnvrgError):
            registry.update(title=1)

        with pytest.raises(CnvrgError):
            registry.update(username=1)



import pytest

from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.workflows import ExperimentsClient


class TestExperimentsClient:

    def test_experiment_with_csv_injection_in_title_error(self, empty_project):
        experiments_client = ExperimentsClient(empty_project)
        malicious_title = "=cmd|’ /C notepad’!‘A1’"

        with pytest.raises(CnvrgArgumentsError):
            experiments_client.create(title=malicious_title)

    def test_experiment_with_csv_injection_in_command_error(self, empty_project):
        experiments_client = ExperimentsClient(empty_project)
        malicious_command = "=cmd|’ /C notepad’!‘A1’"

        with pytest.raises(CnvrgArgumentsError):
            experiments_client.create(command=malicious_command)



import re

from setuptools import find_packages, setup

# regex returns [('version', '0.1.1')] which is then converted to {'version': '0.1.1'}
metadata = dict(re.findall("__([a-z]+)__ = '([^']+)'", open('cnvrgv2/_version.py').read()))
with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='cnvrgv2',
    version=metadata["version"],
    author="Cnvrg",
    author_email="support@cnvrg.io",
    description="Python SDK library for using cnvrg",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        required
    ],
    extras_require={
        'dev': [
            'pytest',
            'pytest-mock'
        ],
        'azure': [
            'azure-core>=1.10.0',
            'azure-storage-blob>=12.10.0'
        ],
        'google': [
            'google-api-core>=1.23.0',
            'google-cloud-core>=1.4.4',
            'google-cloud-storage>=1.32.0',
            'google-auth>=1.23.0',
            'google-crc32c>=1.0.0',
            'google-resumable-media>=1.1.0',
            'googleapis-common-protos>=1.52.0'
        ],
        'python3.6': [  # Remove Feb-23
            'boto3==1.19.12',
            'botocore==1.22.12'
        ]
    },
    entry_points='''
        [console_scripts]
        cnvrgv2=cnvrgv2.cli.cli:safe_entry_point
    ''',
)



from cnvrgv2.modules.labels.utils import LabelColor

# COMMON
CLI_UNEXPECTED_ERROR = "Unexpected error while executing command. details: {0}"
OUTPUT_DIR_LOCATION = "the location of the output directory"

# LOGIN
LOGIN_PROMPT_DOMAIN = "Please enter your cnvrg domain"
LOGIN_PROMPT_EMAIL = "Please enter your email"
LOGIN_PROMPT_PASSWORD = "Please enter your password"
LOGIN_HELP_AUTH_TOKEN = "Enter authentication token instead of password"
LOGIN_ORGANIZATION_HELP = "Organization to log in to"
LOGIN_ALREADY_LOGGED_IN = "Seems you\'re already logged in"
LOGIN_SUCCESS = "Successfully logged in as {0}"
LOGIN_INVALID_CREDENTIALS = "Invalid credentials"

# LOGOUT
LOGOUT_SUCCESS = "Logged out successfully"
LOGOUT_CONFIG_MISSING = "Cannot logout. Config file is missing. Try logging in first"

# ME
ME_SUCCESS_API = "API: {0}/api/v2"
ME_SUCCESS_LOGS = "Logs file located at: {0}"
ME_SUCCESS = "Logged in as: {0}"
ME_LOGGER_SUCCESS = "Successfully printed user's details (cnvrgv2 me)"

# DATA OWNER
DATA_UPLOAD_SUCCESS = "Successfully uploaded updated files"
DATA_DOWNLOAD_SUCCESS = "Successfully downloaded updated files"
DATA_COMMIT_MESSAGE = "Commit message"
DATA_UPLOAD_HELP_GIT_DIFF = "From the list of files, upload those who returned from git diff command."
DATA_UPLOAD_HELP_FORCE = "Enable to create an empty commit without files from parent commit."
DATA_UPLOAD_HELP_OVERRIDE = "Enable to re-upload even if the file already exists."
DATA_LOCATION_IN_STORAGE = "Specify the destination directory for saving uploaded files on cnvrg"
DATA_UPLOAD_HELP_THREADS = "Number of threads that will be used to preform the upload"
DATA_UPLOAD_HELP_CHUNKS = "Number of chunks that will be used to preform the upload"
DATA_PROMPT_COMMIT = "Please enter sha1 to clone"
DATA_PROMPT_CACHE = "Please enter sha1 to cache"
DATA_DOWNLOAD_ERROR = "Unexpected error while executing command"
SYNC_OUTPUT_DIR = "The directory that will be synced inside the project"
DEBUG_MODE_HELP = "Enable to create start commit in debug mode"
CLONE_NUMBER_OF_THREADS = "Number of threads that will be used to preform the clone"
QUERY_SLUG = "Query slug"

# DATASET
DATASET_PROMPT_CLONE = "Please enter dataset name to clone"
DATASET_HELP_CLONE = "Name of the dataset to clone"
DATASET_HELP_CACHE = "Name of the dataset to cache"
DATASET_HELP_UNCACHE = "Name of the dataset to uncache"
DATASET_HELP_CLONE_OVERRIDE = "Whether or not re-clone in case the dataset already cloned"
DATASET_HELP_CLONE_COMMIT = "The commit sha1 to clone"
DATASET_CLONE_SKIP = "Dataset {0} is already cloned, therefore skip clone." \
                     " If you want to override, run again using -o flag."
DATASET_PROMPT_NAME = "Please enter dataset name"
DATASET_PROMPT_DISK_NAME = "Please enter disk name"
DATASET_HELP_NAME = "Name of the dataset"
DATASET_CLONE_SUCCESS = "Successfully cloned dataset: {0}"
DATASET_PUT_PROMPT_FILES = "Please enter a comma separated list of file paths to upload. use . " \
                           "to upload the whole directory"
DATASET_PUT_HELP_FILES = "A comma separated list of file paths to upload. use . " \
                         "to upload the whole directory"
DATASET_REMOVE_PROMPT_FILES = "Please enter a comma separated list of file paths to remove. Wildcards allowed"
DATASET_REMOVE_HELP_FILES = "A comma separated list of file paths to remove. use Wildcards allowed"
DATASET_REMOVE_SUCCESS = "Files removed successfully"
DATASET_PROMPT_DELETE = "Please enter dataset name to delete"
DATASET_HELP_DELETE = "Name of the dataset to delete"
DATASET_HELP_LABELS__NAME = "Name of the dataset."
DATASET_HELP_LABELS__LABEL_NAME = "Name of the label."
DATASET_HELP_LIST_FILES = "Name for the dataset"
DATASET_HELP_LIST_FILES_LENGTH = "Number of files to display"
DATASET_DELETE_CONFIRM = "Are you sure you want to delete dataset: {0}?"
DATASET_DELETE_SUCCESS = "Successfully deleted dataset: {0}"
DATASET_ADD_LABEL_SUCCESS = "Successfully assigned label {0} to dataset: {1}"
DATASET_REMOVE_LABEL_SUCCESS = "Successfully removed label from dataset: {0}"
DATASET_SYNC_SUCCESS = "Successfully synced dataset {0}"
DATASET_PROMPT_CREATE = "Please enter dateset name to create"
DATASET_HELP_CREATE = "Name for the new dataset"
DATASET_CREATE_FOLDER_NOT_EMPTY = "Warning! You're about to associate a non empty folder with the new dataset." \
                                  "\r\nContinue?"
DATASET_CREATING_MESSAGE = "Creating new dataset"
DATASET_CONFIGURING_FOLDER = "Configuring dataset folder"
DATASET_CREATE_SUCCESS = "Successfully created dataset:" \
                         "\r\nDataset name: {0}" \
                         "\r\nDataset slug: {1}"
DATASET_SCAN_START = "Scanning datasets"
DATASET_LIST_FILES_HELP_CHUNK_SIZE = "Number of files to show in each print"
DATASET_LIST_FILES_HELP_LIMIT = "Limit the number of files"
DATASET_SCAN_NO_RESULTS = "No datasets found in working dir"
DATASET_VERIFY_STATUS = "Success: {0}"
DATASET_UNCACHE_SUCCESS = "Dataset successfully uncached"
DATASET_CACHE_SUCCESS = "Dataset successfully cached"
DATASET_NAMES = "Dataset names"
DATASET_NAMES_HELP = "Dataset names to verify"
DATASET_VERIFY_TIMEOUT = "Timeout"

# PROJECT
PROJECT_PROMPT_CLONE = "Please enter project name to clone"
PROJECT_HELP_CLONE = "Name of the project to clone"
PROJECT_PROMPT_DELETE = "Please enter project name to delete"
PROJECT_HELP_DELETE = "Name of the project to delete"
PROJECT_HELP_LIST_FILES = "Name of the project"
PROJECT_HELP_LABELS__NAME = "Name of the project"
PROJECT_HELP_LABELS__LABEL_NAME = "Name of the label."
PROJECT_HELP_CLONE_OVERRIDE = "Whether or not re-clone in case the project already cloned"
PROJECT_HELP_CLONE_COMMIT = "The commit sha1 to clone"
PROJECT_CLONE_SKIP = "Project {0} is already cloned, therefore skip clone." \
                     " If you want to override, run again using -o flag."
PROJECT_CLONE_SUCCESS = "Successfully cloned project: {0}"
PROJECT_DELETE_SUCCESS = "Successfully deleted project: {0}"
PROJECT_ADD_LABEL_SUCCESS = "Successfully assigned label {0} to project: {1}"
PROJECT_REMOVE_LABEL_SUCCESS = "Successfully removed label from project: {0}"
PROJECT_DELETE_CONFIRM = "Are you sure you want to delete project: {0}?"
PROJECT_DOWNLOAD_SUCCESS = "Successfully downloaded updated files"
PROJECT_LINK_PROMPT_NAME = "Please enter a name for the new project"
PROJECT_LINK_HELP_NAME = "Name for the new project"
PROJECT_PROMPT_NAME = "Please enter project name"
PROJECT_HELP_NAME = "Name of the project"
PROJECT_LINK_GIT_PROMPT_NAME = "Please enter the name of the git project"
PROJECT_LINK_GIT_HELP_NAME = "Name of the git project"
SOFT_LINK_GIT_PROJECT = "Link the git project, if it hasn't done before"
PROJECT_GIT_LINK_SKIP = "Project {0} is already linked, therefore skip linking"
PROJECT_JOB_SLUG_HELP_NAME = "Slug of the job"
PROJECT_SLUG_HELP_TITLE = "The slug of the project"
PROJECT_LINK_SUCCESS = "Finished linking project {0} successfully."
PROJECT_CREATE_NEW = "Creating new project {0}"
PROJECT_CONFIGURING_FOLDER = "Configuring project folder"
PROJECT_UPLOAD = "Uploading project files"
PROJECT_PROMPT_CREATE = "Please enter project name to create"
PROJECT_HELP_CREATE = "Name for the new project"
PROJECT_CREATE_FOLDER_NOT_EMPTY = "Warning! You're about to associate a non empty folder with the new project." \
                                  "\r\nContinue?"
PROJECT_CREATING_PROJECT = "Creating new project {0}"
PROJECT_CREATE_SUCCESS = "Successfully created project {0}"
PROJECT_SYNC_SUCCESS = "Successfully synced project {0}, commit_sha1: {1}"
PROJECT_IS_UP_TO_DATE = "\N{check mark} Project {0} is up to date"
PROJECT_HELP_CURRENT_DIR = "Wheteher or not to clone into the current directory"

# FILES
FILES_UPLOAD_SUCCESS = "Successfully uploaded files"
COMMIT_SHA1_MESSAGE = "Commit sha1: {0}"

# LIBRARIES
LIBRARY_PROMPT_CLONE = "Please enter library name to clone"
LIBRARY_VERSION_PROMPT_CLONE = "Please enter library version name to clone (default = \"latest\")"
LIBRARY_CLONE_SKIP = "Library {0} is already cloned, skipping action"
LIBRARY_CLONE_SUCCESS = "Successfully cloned library: {0}"

# MEMBERS
MEMBER_ADDED_SUCCESS = "{0} was added successfully with role {1}"
MEMBER_UPDATED_SUCCESS = "{0} was updated successfully to role {1}"
MEMBER_ENTER_EMAIL = "Please enter a valid user email"
MEMBER_HELP_EMAIL = "A valid email of exists user"
MEMBER_ENTER_ROLE = "Please enter a valid role [admin, manager, member, reviewer]"
MEMBER_HELP_ROLE = "A role for the user, must be one of the following: admin, manager, member, reviewer"
MEMBER_REMOVED_SUCCESS = "{0} was removed successfully"

# EXPERIMENT
EXPERIMENT_PROMPT_TITLE = "Please enter a title for the experiment"
EXPERIMENT_PROMPT_COMMIT = "Please enter commit to merge"
EXPERIMENT_ARTIFACTS_PROMPT_COMMIT = "Please enter commit"
EXPERIMENT_HELP_COMMIT = "Commit sha1 to merge"
EXPERIMENT_ARTIFACTS_HELP_COMMIT = "Commit sha1 of artifacts"
EXPERIMENT_HELP_TITLE = "Name of the experiment"
EXPERIMENT_HELP_TEMPLATES = "Comma separated list of template names"
EXPERIMENT_HELP_LOCAL = "Boolean. Run the experiment locally"
EXPERIMENT_PROMPT_COMMAND = "Please enter a command to run as experiment"
EXPERIMENT_HELP_COMMAND = "The command to run"
EXPERIMENT_HELP_DATASETS = "List of dictionaries containing slug, commit sha1 or query"
EXPERIMENT_HELP_DATASOURCES = "List of datasource slugs"
EXPERIMENT_HELP_VOLUME = "A volume name to attach to this experiment"
EXPERIMENT_HELP_SYNC_BEFORE = "Boolean. Sync environment before running the experiment"
EXPERIMENT_HELP_SYNC_AFTER = "Boolean. sync environment after the experiment finished"
EXPERIMENT_HELP_IMAGE = "Image name and tag to create experiment with. format - image_name:tag"
EXPERIMENT_HELP_GIT_BRANCH = "The branch to pull files from for the experiment, in case project is git project"
EXPERIMENT_HELP_GIT_COMMIT = "The commit to pull files from for the experiment, in case project is git project"
EXPERIMENT_HELP_GRID = "Run a grid search using a YAML file outlining the parameters"
EXPERIMENT_HELP_LOCAL_FOLDERS = "Local folders to mount with experiment"
EXPERIMENT_CREATE_SUCCESS = "Experiment {0} created successfully. \r\nExperiment is available at: {1}"
EXPERIMENT_GRID_CREATE_SUCCESS = "Grid {0} created successfully. \r\nGrid is available at: {1}"
EXPERIMENT_MERGE_SUCCESS = "Commit successfully merged to master"
EXPERIMENT_GIT_ERROR_MESSAGE = "Cannot merge commits for git projects"
EXPERIMENT_DATASET_QUERY_ERROR_MESSAGE = "Cant use commit and query for dataset {0}, please use only one"
EXPERIMENT_DOES_NOT_EXIST = "Couldn't find experiment. Try with env variables or passing --slug"
EXPERIMENT_LOG_ARTIFACTS_SUCCESS = "Successfully logged artifacts"
EXPERIMENT_LOG_IMAGES_SUCCESS = "Successfully logged images"
EXPERIMENT_PULL_ARTIFACTS_SUCCESS = "Successfully pulled artifacts"
EXPERIMENT_HELP_WORK_DIR = "Working Dir to upload files to"
EXPERIMENT_HELP_OUTPUT_DIR = "Sets the output folder for cnvrg to track experiment artifacts"
EXPERIMENT_DELETE_ARTIFACTS = "Delete related artifacts from the storage"
EXPERIMENT_DELETE_SUCCESS = "Successfully deleted experiment"
EXPERIMENT_DELETE_ARTIFACTS_PROMPT = "This action will delete related artifacts from the storage. Are you sure?"
EXPERIMENT_DELETE_ARTIFACTS_ABORTED = "User aborted delete experiments"
EXPERIMENT_HELP_LOG = "Display the log of the experiment"
EXPERIMENT_HELP_WAIT = "Wait for the experiment to finish"

# LOGS
LOG_START_COMMAND = "Starting command {0}. Options: {1}"
LOG_CLONING_PROJECT = "Cloning project: {0}"
LOG_CLONING_LIBRARY = "Cloning library: {0}=={1}"
LOG_CLONING_DATASET = "Cloning dataset: {0}"
LOG_LIST_FILES_PROJECT = "List files from project: {0}"
LOG_LIST_FILES_DATASET = "List files from dataset: {0}"
LOG_LIST_LABELS_DATASET = "List labels from dataset: {0}"
LOG_LIST_LABELS_PROJECT = "List labels from project: {0}"


# SSH
SSH_HELP_PORT = "Port to bind on host"
SSH_HELP_USERNAME = "Username to login in container, default will be image default user"
SSH_HELP_PASSWORD = "Password for login"
SSH_HELP_KUBECTL = "Full path to kubeconfig file, otherwise default will be used"
SSH_STARTING_SESSION = "Starting a new ssh session"
SSH_WAITING_FOR_READY = "Waiting for ssh session to be ready..."
SSH_READY = "\r\nSsh session is ready to receive connections.\r\n" \
            "\r\nIn order to connect to your job, define your ssh connection with the following params:\r\n" \
            "host: 127.0.0.1\r\n" \
            "port: {0}\r\n" \
            "username: {1}\r\n" \
            "password: {2}"

# WORKSPACE
WORKSPACE_DOES_NOT_EXIST = "Workspace was not found, check env or pass slug"
WORKSPACE_SLUG = "Slug of the workspace"
WORKSPACE_HELP_SYNC_REMOTE = "Sync the workspace"
WORKSPACE_HELP_DATASETS = "List of comma separated datasets names to use in the workspace"
WORKSPACE_HELP_DATASOURCE = "List datasource slugs"
WORKSPACE_HELP_VOLUME = "A volume name to attach to this workspace"
WORKSPACE_HELP_GIT_BRANCH = "The branch to pull files from for the workspace, in case project is git project"
WORKSPACE_HELP_GIT_COMMIT = "The commit to pull files from for the workspace, in case project is git project"
WORKSPACE_HELP_TEMPLATES = "Comma separated list of template names"
WORKSPACE_HELP_NOTEBOOK_TYPE = "The type of the created notebook"
WORKSPACE_HELP_TITLE = "Name of the workspace"
WORKSPACE_HELP_IMAGE = "Image name and tag to create workspace with. format - image_name:tag"
WORKSPACE_CREATE_SUCCESS = "Workspace {0} created successfully. \r\nWorkspace is available at: {1}"
WORKSPACE_HELP_COMMIT = "Sha1 of the commit to use, default: latest"
WORKSPACE_HELP_LOCAL_FOLDERS = "Local folders to mount with workspace"

# DATASOURCE
DATASOURCE_SLUG_HELP = "The slug of the requested datasource"
DATASOURCE_SLUG = "The slug of the datasource"
DATASOURCE_PAGE_SIZE_HELP = "Page size: The size of each individual page downloaded"
DATASOURCE_MAX_WORKERS_HELP = "Max number of workers that clone the datastore at the same time"
DATASOURCE_LOCAL_DIRECTORY_HELP = "When true, The files will be downloaded directly to the current local directory"
DATASOURCE_NAME = "The given name of the datasource"
DATASOURCE_NAME_HELP = "Provide a unique name for the datasource you are creating"
DATASOURCE_TYPE = "The storage type"
DATASOURCE_TYPE_HELP = "Storage type (like: s3...)"
DATASTORE_BUCKET = "The name of the bucket associated with the datasource"
DATASTORE_BUCKET_HELP = "Name the bucket you wish to be associated with the datasource"
DATASOURCE_CREDENTIALS = "Credentials to the datasource"
DATASOURCE_CREDENTIALS_HELP = "Credentials to the datasource in form of a dictionary"
DATASOURCE_REGION = "The region of the bucket associated with the datasource"
DATASOURCE_REGION_HELP = "The region of the bucket associated with the datasource"
DATASOURCE_DESCRIPTION = "Description"
DATASOURCE_DESCRIPTION_HELP = "Description"
DATASOURCE_PUBLIC = "Is the datasource public?"
DATASOURCE_PUBLIC_HELP = "Is the datasource public?"
DATASOURCE_PATH = "The Path for items within the datasource"
DATASOURCE_PATH_HELP = "The Path for items within the datasource"
DATASOURCE_ENDPOINT = "The endpoint of the service associated with the datasource"
DATASOURCE_ENDPOINT_HELP = "The endpoint of the service associated with the datasource"
DATASOURCE_USERS = "The users with permissions to the datasource"
DATASOURCE_USERS_HELP = "The users with permissions to the datasource (a list of emails)"
DATASOURCE_FORCE = "Force re-cloning datasource. Note: All previous local data will be removed!"
DATASOURCE_IGNORE_FILE_ERROR = "Ignore the cloning request of the datasource if it already exists"

# WEBAPP
WEBAPP_HELP_SLUG = "Slug of the webapp"
WEBAPP_HELP_FREQUENCY = "The frequency to pull the logs event in seconds"


# ENDPOINT
ENDPOINT_DOES_NOT_EXIST = "Endpoint was not found, check env or pass slug"
ENDPOINT_SLUG = "Slug of the endpoint"
ENDPOINT_HELP_VERSION_SLUG = "Slug of the version to rollback to"
ENDPOINT_PROMPT_METRIC_NAME = "Please enter the metric name"
ENDPOINT_PROMPT_METRIC_X = "Please enter the metric x value"
ENDPOINT_PROMPT_METRIC_Y = "Please enter the metric y value"
ENDPOINT_PROMPT_FUNCTION_NAME = "Please enter The name of the function the endpoint will route to"
ENDPOINT_PROMPT_FILE_NAME = "Please enter the file name"
ENDPOINT_PROMPT_TITLE_NAME = "Please enter the endpoint title name"
ENDPOINT_METRIC_NAME = "The metric name"
ENDPOINT_METRIC_X = "The metric x value"
ENDPOINT_METRIC_Y = "The metric y value"
ENDPOINT_ROLLBACK_SUCCESS = "Endpoint {0} rollback to version slug {1} successfully"
ENDPOINT_UPDATE_SUCCESS = "Endpoint {0} updated successfully"
ENDPOINT_HELP_ARGS = "Optional arguments"
ENDPOINT_HELP_MAX_TIMEOUT = "Time in minutes to wait until timeout"
ENDPOINT_HELP_PREP_FUNCTION = "Name of new preprocess function"
ENDPOINT_HELP_PREP_FILE = "Name of new preprocess file"
ENDPOINT_HELP_FUNCTION_NAME = "The name of the function the endpoint will route to"
ENDPOINT_HELP_FILE_NAME = "The file containing the endpoint's functions"
ENDPOINT_UPDATE_IN_PROGRESS = "Updating endpoint..."
ENDPOINT_HELP_TITLE = "Name of the endpoint"
ENDPOINT_HELP_KIND = "The endpoint kind"
ENDPOINT_HELP_ENV_SETUP = "The interpreter to use python_2,python_3,pyspark,r_endpoint"
ENDPOINT_HELP_TEMPLATES = "Comma separated list of template names"
ENDPOINT_HELP_KAFKA_BROKERS = "Comma separated list of kafka brokers"
ENDPOINT_HELP_KAFKA_INPUT_TOPICS = "Comma separated list of kafka input topics"
ENDPOINT_CREATE_SUCCESS = "Endpoint {0} created successfully. \r\nEndpoint is available at: {1}"
ENDPOINT_PROMPT_LOG_MESSAGE = "Please enter the log messages, separated by comma"
ENDPOINT_HELP_LOG_MESSAGE = "Comma separated list of logs to write to endpoint"
ENDPOINT_HELP_LOGS_TYPE = "Level of the logs , default: output. [output, error, warning, info]"
ENDPOINT_LOG_LEVEL_NOT_SUPPORTED = "Log level {0} not supported, supported levels: output, error, warning, info"
ENDPOINT_NO_PREDICTIONS = "No predictions were found for endpoint"
ENDPOINT_HELP_MODEL_NUMBER = "The endpoint model number to query"
ENDPOINT_HELP_BATCH_SIZE = "The batch size of predictions to return"
ENDPOINT_HELP_OFFSET = "The offset of the predictions to return"
ENDPOINT_HELP_END_TIME = "The end time of the predictions to return"
ENDPOINT_HELP_START_TIME = "The start time of the predictions to return"

# FLOWS
FLOW_YAML_PATH = "Path to yaml file, describing the new flow"
FLOW_CREATE_SUCCESS = "Flow {0} created successfully."

# CONFIG
CONFIG_HELP_CHECK_CERTIFICATE = "{0} ssl validation on https requests"
CONFIG_UPDATE_SUCCESS = "Config updated successfully"
CONFIG_HELP_ORGANIZTION = "Name of organization to switch to"
CONFIG_NO_ARGS_LOG = "No arguments sent. Showing help message instead"

# IMAGE
IMAGE_HELP_TAG = "Image tag"
IMAGE_HELP_SLUG = "Image slug"
IMAGE_HELP_NAME = "Image repository name"
IMAGE_HELP_LOGO = "Image logo"
IMAGE_HELP_CUSTOM = "Is custom image (requires dockerfile)"
IMAGE_HELP_README = "Readme path for the image"
IMAGE_HELP_REGISTRY = "Image registry"
IMAGE_HELP_DOCKERFILE = "Dockerfile path to build a custom image"

IMAGE_PROMPT_TAG = "Please enter image tag"
IMAGE_PROMPT_SLUG = "Please enter image slug"
IMAGE_PROMPT_NAME = "Please enter image name (repository)"
IMAGE_PROMPT_LOGO = "Please enter image logo"
IMAGE_PROMPT_REGISTRY = "Please enter image registry"

IMAGE_INVALID_README = "Readme path is invalid"
IMAGE_INVALID_DOCKERFILE = "Dockerfile path is invalid"

IMAGE_CREATE_SUCCESS = "Successfully created image {0}"

# REGISTRY
REGISTRY_HELP_URL = "Registry url"
REGISTRY_HELP_TYPE = "Registry type (cnvrg, dockerhub, gcr, acr, ecr, ...)"
REGISTRY_HELP_SLUG = "Registry slug"
REGISTRY_HELP_TITLE = "Registry title"
REGISTRY_HELP_USERNAME = "Registry username for private registries"
REGISTRY_HELP_PASSWORD = "Registry password for private registries"

REGISTRY_PROMPT_URL = "Please enter registry url"
REGISTRY_PROMPT_TITLE = "Please enter registry tag"

REGISTRY_CREATE_SUCCESS = "Successfully created registry {0}"

# LABELS
LABEL_HELP_KIND = "Label kind (projects, datasets)"
LABEL_HELP_NAME = "Label name"
LABEL_HELP_COLOR_NAME = "Label color. available colors: {0}".format(LabelColor.ALL_COLORS)

LABEL_PROMPT_KIND = "Please enter the kind (either 'projects' or 'datasets')"

LABEL_CREATE_SUCCESS = "Successfully created label {0}"
LABEL_DELETE_SUCCESS = "Successfully deleted dataset: {0}"

# ORGANIZATION SETTINGS
ORGANIZATION_SETTINGS_HELP_DEFAULT_COMPUTES = "Default computes"
ORGANIZATION_SETTINGS_HELP_INSTALL_DEPENDENCIES = "Install dependencies"
ORGANIZATION_SETTINGS_HELP_SLACK_WEBHOOK_URL = "Slack webhook URL"
ORGANIZATION_SETTINGS_HELP_DEBUG_TIME = "Debug time"
ORGANIZATION_SETTINGS_HELP_EMAIL_ON_ERROR = "Send email on error"
ORGANIZATION_SETTINGS_HELP_EMAIL_ON_SUCCESS = "Send email on success"
ORGANIZATION_SETTINGS_HELP_QUEUED_COMPUTE_WAIT_TIME = "Queued compute wait time"
ORGANIZATION_SETTINGS_HELP_IDLE_ENABLED = "Enable or Disable idle time"
ORGANIZATION_SETTINGS_HELP_IDLE_TIME = "Idle time"
ORGANIZATION_SETTINGS_HELP_MAX_DURATION_WORKSPACES = "Max duration for workspaces"
ORGANIZATION_SETTINGS_HELP_MAX_DURATION_EXPERIMENTS = "Max duration for experiments"
ORGANIZATION_SETTINGS_HELP_MAX_DURATION_ENDPOINTS = "Max duration for endpoints"
ORGANIZATION_SETTINGS_HELP_MAX_DURATION_WEBAPPS = "Max duration for webapps"
ORGANIZATION_SETTINGS_HELP_AUTOMATICALLY_CLEAR_CACHED_COMMITS = "Number of cached commits to be cleared" \
                                                                " automatically"
ORGANIZATION_SETTINGS_HELP_CUSTOM_PYPI_ENABLED = "Enable custom PYPI"
ORGANIZATION_SETTINGS_HELP_CUSTOM_PYPI_URL = "Custom PYPI URL"



import shutil
import time
from datetime import datetime
from cnvrgv2.config import error_messages
from dateutil.tz import tzlocal
from cnvrgv2.errors import CnvrgAlreadyClonedError


def handle_dir_exist(force: bool, path: str) -> None:
    """
    Handle the case where a directory already exists.

    @param force: If True, forcefully delete the directory.
    @type force: bool
    @param path: The path to the directory.
    @type path: str
    @raise CnvrgAlreadyClonedError: If the directory already exists and force is False.
    """
    if force:
        shutil.rmtree(path)
    else:
        raise CnvrgAlreadyClonedError(error_messages.DATASOURCE_ALREADY_CLONED)


def format_credentials(creds: dict) -> dict:
    """
    Format the credentials for use with boto3.

    @param creds: The raw credentials.
    @type creds: dict
    @return: The formatted credentials.
    @rtype: dict
    """
    expiration = creds.get('expiration', 129600)
    return {
        "access_key": creds['access_key_id'],
        "secret_key": creds['secret_access_key'],
        "token": creds.get('session_token'),
        "expiry_time": datetime.fromtimestamp(time.time() + expiration, tz=tzlocal()).isoformat(),
    }


def get_refreshable_credentials_function(datasource):
    """
    Get a function to refresh credentials for the given data source.

    @param datasource: The data source instance.
    @type datasource: Datasource
    @return: A function to refresh credentials.
    @rtype: function
    """

    def refresh():
        creds = datasource._get_credentials()
        return format_credentials(creds)

    return refresh



from datetime import timedelta, datetime

from dateutil.relativedelta import relativedelta


def get_relative_time(days=None, months=None, hours=None):
    """
    Get relative time
    @param days: Number of days
    @param months: Number of months
    @param hours: Number of hours
    @return: Time
    """

    now = datetime.now()
    if days:
        return (now - timedelta(days=days)).strftime("%Y-%m-%d %H:%M:%S %z")
    if months:
        return (now - relativedelta(months=months)).strftime("%Y-%m-%d %H:%M:%S %z")
    if hours:
        return (now - timedelta(hours=hours)).strftime("%Y-%m-%d %H:%M:%S %z")
    return now.strftime("%Y-%m-%d %H:%M:%S %z")



import click

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command


@click.group(name='webapp')
def webapp_group():
    pass


@webapp_group.command()
@click.option('-s', '--slug', default=None, help=messages.WEBAPP_HELP_SLUG)
@click.option('-f', '--frequency', default=5, help=messages.WEBAPP_HELP_FREQUENCY)
@click.option('-ps', '--project-slug', help=messages.PROJECT_SLUG_HELP_TITLE, default=None, required=False)
@prepare_command()
def compare_experiments(webapp, logger, slug, frequency, project_slug):
    """
      start a compare experiments in webapp for tensorboard
    """
    logger.log_and_echo("Compare is running")
    webapp.compare_experiments(frequency=frequency)



import pytest

from cnvrgv2.errors import CnvrgArgumentsError, CnvrgFileError
from cnvrgv2.modules.dataset import Dataset
from cnvrgv2.modules.flows.flow import Flow
from cnvrgv2.modules.flows.flows_client import FlowsClient
from cnvrgv2.proxy import Proxy


class TestFlows:

    def test_flow_toggle_webhook_success(self, mocker, empty_context):
        flow = Flow(context=empty_context, slug="flow-slug")
        mocked_toggle = mocker.patch.object(Proxy, 'call_api')
        flow.toggle_webhook(True)
        assert mocked_toggle.call_count == 1

    def test_flow_toggle_webhook_wrong_argument_type(self, empty_context):
        flow = Flow(context=empty_context, slug="flow-slug")
        with pytest.raises(CnvrgArgumentsError):
            flow.toggle_webhook("on")

    def test_flow_toggle_dataset_update_success(self, mocker, empty_context):
        flow = Flow(context=empty_context, slug="flow-slug")
        dataset = Dataset(context=empty_context, slug="dataset-slug")
        mocked_toggle = mocker.patch.object(Proxy, 'call_api')
        flow.toggle_dataset_update(True, dataset=dataset)
        assert mocked_toggle.call_count == 1

    def test_flow_toggle_dataset_update_wrong_toggle_argument_type(self, empty_context):
        flow = Flow(context=empty_context, slug="flow-slug")
        with pytest.raises(CnvrgArgumentsError):
            flow.toggle_dataset_update("on")

    def test_flow_toggle_dataset_update_wrong_dataset_argument_type(self, empty_context):
        flow = Flow(context=empty_context, slug="flow-slug")
        with pytest.raises(CnvrgArgumentsError):
            flow.toggle_dataset_update(True, "dataset")

    def test_flow_toggle_dataset_update_wrong_toggle_and_dataset_argument_types(self, empty_context):
        flow = Flow(context=empty_context, slug="flow-slug")
        with pytest.raises(CnvrgArgumentsError):
            flow.toggle_dataset_update("on", "dataset")

    def test_flow_set_schedule_value_success(self, mocker, empty_context):
        flow = Flow(context=empty_context, slug="flow-slug")
        mocked_toggle = mocker.patch.object(Proxy, 'call_api')
        flow.set_schedule("59 12 25 3 4")
        assert mocked_toggle.call_count == 1

    def test_flow_set_illegal_schedule_value_fail(self, empty_context):
        flow = Flow(context=empty_context, slug="flow-slug")
        with pytest.raises(CnvrgArgumentsError):
            flow.set_schedule("1000 1000 * * *")

    def test_flow_set_insufficient_schedule_arguments_fail(self, empty_context):
        flow = Flow(context=empty_context, slug="flow-slug")
        with pytest.raises(CnvrgArgumentsError):
            flow.set_schedule("20 10")

    def test_flow_create_return_error_wrong_yaml_path_fail(self, empty_context):
        with pytest.raises(CnvrgFileError):
            FlowsClient.create(self, yaml_path="bad/yaml/path")

    def test_flow_create_return_error_wrong_yaml_string_fail(self, empty_context):
        with pytest.raises(CnvrgArgumentsError):
            FlowsClient.create(self, yaml_string={})

    def test_flow_create_return_error_both_wrong_yaml_params_fail(self, empty_context):
        with pytest.raises(CnvrgArgumentsError):
            FlowsClient.create(self, yaml_string="string", yaml_path='/path')



from tests.e2e.data_owner_e2e.data_owner_files import DataOwnerFiles


class TestProjectFiles(DataOwnerFiles):

    def test_put_files_success(self, e2e_project, temp_file):
        self.data_owner_put_files(e2e_project, temp_file)

    def test_put_unchanged_files_success(self, e2e_project, temp_file):
        self.data_owner_put_unchanged_files(e2e_project, temp_file)

    def test_put_folder_success(self, e2e_project, tmpdir):
        self.data_owner_put_folder(e2e_project, tmpdir)

    def test_put_files_override_false_success(self, e2e_project, temp_file):
        self.data_owner_put_files_override_false(e2e_project, temp_file)

    def test_put_files_override_true_success(self, e2e_project, temp_file):
        self.data_owner_put_files_override_true(e2e_project, temp_file)

    def test_put_file_abs_path_dir_path_success(self, e2e_project, temp_file):
        self.data_owner_put_files_abs_path_dir_path(e2e_project, temp_file)

    def test_put_file_dir_path_relative_path_success(self, e2e_project, temp_file):
        self.data_owner_put_file_dir_path_relative_path(e2e_project, temp_file)

    def test_put_folder_abs_path_dir_path_success(self, e2e_project, tmpdir):
        self.data_owner_put_folder_abs_path_dir_path(e2e_project, tmpdir)

    def test_put_folder_dir_path_relative_path_success(self, e2e_project, tmpdir):
        self.data_owner_put_folder_dir_path_relative_path(e2e_project, tmpdir)

    def test_put_empty_folder_dir_path_relative_path_success(self, e2e_project, tmpdir):
        self.data_owner_put_empty_folder_dir_path_relative_path(e2e_project, tmpdir)

    def test_put_files_force_success(self, e2e_project, temp_file, temp_file_second):
        self.data_owner_put_files_force(e2e_project, temp_file, temp_file_second)

    def test_list_files_success(self, e2e_project, tmpdir):
        self.data_owner_list_files(e2e_project, tmpdir)

    def test_list_files_with_names_success(self, e2e_project, tmpdir):
        for num_files in [5, 20, 30, 60]:
            self.data_owner_list_files_var(e2e_project, tmpdir, num_files)

    def test_list_files_and_folders_success(self, e2e_project, tmpdir):
        self.data_owner_list_files_and_folders(e2e_project, tmpdir)

    def test_clone_files_override_success(self, e2e_project, tmpdir, temp_file):
        self.data_owner_clone_override(e2e_project, tmpdir, temp_file)

    def test_delete_files_success(self, e2e_project, temp_file):
        self.data_owner_delete_files(e2e_project, temp_file)

    def test_clone_files_success(self, e2e_project, tmpdir):
        self.data_owner_clone_files(e2e_project, tmpdir)

    def test_clone_with_empty_folder_success(self, e2e_project, tmpdir):
        self.data_owner_clone_with_empty_folder(e2e_project, tmpdir)

    def test_upload_files_success(self, e2e_project, tmpdir):
        self.data_owner_upload_files(e2e_project, tmpdir)

    def test_upload_unchanged_files_success(self, e2e_project, tmpdir):
        self.data_owner_upload_unchanged_files(e2e_project, tmpdir)

    def test_upload_files_from_non_cnvrg_fails(self, e2e_project, tmpdir):
        self.data_owner_upload_files_from_non_cnvrg(e2e_project, tmpdir)

    def test_download_files_success(self, e2e_project):
        self.data_owner_download_files(e2e_project)

    def test_data_owner_download_empty_folder_success(self, e2e_project, tmpdir):
        self.data_owner_download_empty_folder(e2e_project, tmpdir)

    def test_sync_local_success(self, e2e_project):
        self.data_owner_sync_local(e2e_project)

    def test_sync_local_with_output_success(self, e2e_project):
        self.data_owner_sync_local_with_output_dir(e2e_project)

    def test_put_files_absolute_path_success(self, tmpdir, e2e_project):
        self.data_owner_put_files_absolute_path(tmpdir, e2e_project)

    def test_remove_files_absolute_path_success(self, tmpdir, e2e_project):
        self.data_owner_remove_files_absolute_path(tmpdir, e2e_project)

    def test_remove_folders_through_upload_success(self, e2e_project, tmpdir):
        self.data_owner_remove_folders_through_upload(tmpdir, e2e_project)

    def test_remove_file_through_upload(self, e2e_project, tmpdir):
        self.data_owner_remove_file_through_upload(tmpdir, e2e_project)

    def test_list_commits_success(self, e2e_project, temp_file):
        self.data_owner_list_commits(e2e_project, temp_file)

    def test_filtered_list_commits_success(self, e2e_project, temp_file, second_temp_file):
        first_commit_sha_1 = e2e_project.put_files(paths=[temp_file.strpath], message='commit message 1')
        e2e_project.put_files(paths=[second_temp_file.strpath], message='commit message 2')
        commits_generator = e2e_project.list_commits(filter={'message': 'commit message 1'})

        commits = []

        for commit in commits_generator:
            commits.append(commit.sha1)

        assert [first_commit_sha_1] == commits

    def test_empty_filtered_list_commits(self, e2e_project, temp_file):
        e2e_project.put_files(paths=[temp_file.strpath], message='commit message 1')
        commits_generator = e2e_project.list_commits(filter={'message': 'this commit doesnt exist'})

        assert next(commits_generator, None) is None

    def test_access_all_commit_fields_success(self, e2e_project, temp_file, second_temp_file):
        self.data_owner_access_all_commit_fields(e2e_project, temp_file, second_temp_file)

    def test_remove_folder_locally_through_download_success(self, e2e_project, tmpdir):
        self.data_owner_remove_folder_locally_through_download(tmpdir, e2e_project)

    def test_remove_file_locally_through_download_success(self, e2e_project, tmpdir):
        self.data_owner_remove_file_locally_through_download(tmpdir, e2e_project)

    def test_data_owner_sync_complex(self, e2e_project, tmpdir):
        self.data_owner_sync_complex(tmpdir, e2e_project)

    def test_data_owner_sync_removed_file_success(self, e2e_project, tmpdir):
        self.data_owner_sync_removed_file(e2e_project, tmpdir)

    def test_data_owner_sync_removed_folder_success(self, e2e_project, tmpdir):
        self.data_owner_sync_removed_folder(e2e_project, tmpdir)

    def test_data_owner_verify_cnvrg_files_exists_success(self, e2e_project, tmpdir):
        self.data_owner_verify_cnvrg_files_exists(e2e_project, tmpdir)

    def test_sync_zero_size_files_success(self, e2e_git_project):
        self.data_owner_zero_size_file(e2e_git_project)



import pytest

from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.labels.utils import LabelColor, LabelKind
from tests.e2e.conftest import call_database


class TestLabels:
    @staticmethod
    def cleanup(context_prefix):
        delete_command = "DELETE FROM labels WHERE name LIKE'{}%'".format(context_prefix)
        call_database(delete_command)

    def test_create_label__of_kind_project(self, class_context, e2e_client):
        name = class_context.generate_name(5)
        color_name = LabelColor.GREEN
        kind = LabelKind.PROJECT

        label = e2e_client.labels.create(name=name, kind=kind, color_name=color_name)

        assert label.name == name
        assert label.color_name == color_name
        assert label.kind == kind

    def test_create_label__of_kind_dataset(self, class_context, e2e_client):
        name = class_context.generate_name(5)
        color_name = LabelColor.YELLOW
        kind = LabelKind.DATASET

        label = e2e_client.labels.create(name=name, kind=kind, color_name=color_name)

        assert label.name == name
        assert label.color_name == color_name
        assert label.kind == kind

    def test_get_label(self, e2e_client, e2e_label_project):
        label = e2e_client.labels.get(name=e2e_label_project.name, kind=e2e_label_project.kind)
        assert label.name == e2e_label_project.name
        assert label.color_name == e2e_label_project.color_name

    def test_update_label(self, e2e_client, e2e_label_project):
        label = e2e_client.labels.get(name=e2e_label_project.name, kind=e2e_label_project.kind)
        assert label.name == e2e_label_project.name

        label.update(color_name=LabelColor.YELLOW)

        assert label.color_name == LabelColor.YELLOW

    def test_delete_label(self, e2e_client, e2e_label_project):
        label = e2e_client.labels.get(name=e2e_label_project.name, kind=e2e_label_project.kind)
        assert label.name == e2e_label_project.name

        label.delete()
        with pytest.raises(CnvrgArgumentsError):
            e2e_client.labels.get(name=e2e_label_project.name, kind=e2e_label_project.kind).color_name



from datetime import datetime
from dateutil import tz, parser
from cnvrgv2.config import routes
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.url_utils import urljoin


class FlowVersion(DynamicAttributes):

    available_attributes = {
        "title": str,
        "slug": str,
        "id": int,
        "href": str,
        "state": str,
        "created_at": datetime,
        "start_time": datetime,
        "end_time": datetime,
        "schedule_time": datetime
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        if slug:
            self._context.set_scope(SCOPE.FLOW_VERSION, slug)

        scope = self._context.get_scope(SCOPE.FLOW_VERSION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.FLOW_VERSION_BASE.format(scope["organization"], scope["project"],
                                                      scope["flow"], scope["flow-version"])
        self._attributes = attributes or {}

        self.convert_time()
        self._type = "FlowVersion"
        self.slug = scope["flow-version"]

    def info(self):
        """
        Get info of the current flow version
        @return: info of the flow version
        """
        info_url = urljoin(self._route, routes.FLOW_VERSION_INFO)

        response = self._proxy.call_api(
            route=info_url,
            http_method=HTTP.GET
        )
        return response.attributes

    def stop(self):
        """
        Stops the current flow version, in case it's running
        @return: None
        """
        stop_url = urljoin(self._route, routes.FLOW_VERSION_STOP)

        self._proxy.call_api(
            route=stop_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes={})
        )

    def convert_time(self):
        """
        Internal conversion of schedule time
        """
        if not self._attributes.get("schedule_time", None):
            return

        self._attributes["schedule_time"] = parser.parse(self._attributes["schedule_time"])
        self._attributes["schedule_time"] = self._attributes["schedule_time"].astimezone(tz.tzlocal()).strftime(
            "%Y-%m-%dT%H:%M:%S.%fZ")



import os

import click
from prettytable import PrettyTable

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.cli.utils.helpers import print_generator_in_chunks
from cnvrgv2.config import Config, CONFIG_FOLDER_NAME
from cnvrgv2.config import error_messages
from cnvrgv2.utils.storage_utils import cnvrgignore_exists, create_cnvrgignore
from cnvrgv2.modules.labels.utils import LabelKind


@click.group(name='project')
def project_group():
    pass


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_PROMPT_CREATE, help=messages.PROJECT_HELP_CREATE)
@prepare_command()
def create(cnvrg, logger, name):
    """
    Creates a new project, and associates the current folder with the created project.
    It is recommended to create a new project in an empty folder.
    To create a project from a folder that contains content please refer to project link command.
    """
    if os.listdir():
        click.confirm(messages.PROJECT_CREATE_FOLDER_NOT_EMPTY, default=False, abort=True)

    logger.log_and_echo(messages.PROJECT_CREATING_PROJECT.format(name))
    new_project = cnvrg.projects.create(name)

    new_project.local_commit = new_project.commit

    logger.log_and_echo(messages.PROJECT_CONFIGURING_FOLDER)
    new_project.save_config()

    logger.log_and_echo(messages.PROJECT_CREATE_SUCCESS.format(new_project.title))


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_PROMPT_CLONE, help=messages.PROJECT_PROMPT_CLONE)
@click.option('-o', '--override', is_flag=True, default=False, help=messages.PROJECT_HELP_CLONE_OVERRIDE)
@click.option('-c', '--commit', prompt=messages.DATA_PROMPT_COMMIT, default="latest",
              help=messages.PROJECT_HELP_CLONE_COMMIT)
@click.option('-d', '--current_dir', help=messages.PROJECT_HELP_CURRENT_DIR, is_flag=True, default=False)
@prepare_command()
def clone(cnvrg, logger, name, override, commit, current_dir):
    """
    Clones the given project to local folder
    """
    project = cnvrg.projects.get(name)
    logger.info(messages.LOG_CLONING_PROJECT.format(name))
    config_path = project.slug + '/' + CONFIG_FOLDER_NAME
    if current_dir:
        config_path = CONFIG_FOLDER_NAME
    if os.path.exists(config_path) and not override:
        logger.log_and_echo(messages.PROJECT_CLONE_SKIP.format(name))
        return
    project.clone(progress_bar_enabled=True, override=override, commit=commit, current_dir=current_dir)
    success_message = messages.PROJECT_CLONE_SUCCESS.format(name)
    logger.log_and_echo(success_message)


@project_group.command(name='list-files')
@click.option('-n', '--name', prompt=messages.PROJECT_HELP_LIST_FILES, help=messages.PROJECT_HELP_LIST_FILES)
@click.option('-cs', '--chunk-size', default=20, help=messages.DATASET_LIST_FILES_HELP_CHUNK_SIZE)
@click.option('-l', '--limit', type=int, help=messages.DATASET_LIST_FILES_HELP_LIMIT)
@prepare_command()
def list_files(cnvrg, logger, name, chunk_size, limit):
    """
    List files within a project
    """
    project = cnvrg.projects.get(name)
    files = project.list_files()
    logger.info(messages.LOG_LIST_FILES_PROJECT.format(name))
    attributes_field_name = ["File Name", "File Size", "SHA1"]
    attributes = ["file_name", "file_size", "sha1"]
    print_generator_in_chunks(files, chunk_size, limit, attributes, attributes_field_name, line_numbers=True)


@project_group.command(name='list')
@prepare_command()
def list_command(cnvrg, logger):
    """
    list projects title
    """
    projects = cnvrg.projects.list()
    domain = cnvrg._context.domain
    organization = cnvrg._context.organization
    table = PrettyTable()
    table.field_names = ["Project Title", "Project Link", "Public"]
    table.align["Project Title"] = "l"
    table.align["Project Link"] = "l"
    table.align["Public"] = "l"
    for project in projects:
        url = "{0}/{1}/projects/{2}".format(domain, organization, project.slug)
        table.add_row([project.title, url, project.public])
    click.echo(table)


@project_group.command(name='labels')
@click.option('-n', '--name', prompt=messages.PROJECT_HELP_LABELS__NAME, help=messages.PROJECT_HELP_LABELS__NAME)
@prepare_command()
def labels_command(cnvrg, logger, name):
    """
    List labels within a project
    """
    project = cnvrg.projects.get(name)
    labels = project.labels.list()
    logger.info(messages.LOG_LIST_LABELS_PROJECT.format(name))
    attributes_field_name = ["Name", "Color Name"]
    attributes = ["name", "color_name"]
    print_generator_in_chunks(labels, 20, 10, attributes, attributes_field_name, line_numbers=True)


@project_group.command(name='add-label')
@click.option('-n', '--name', prompt=messages.PROJECT_HELP_LABELS__NAME, help=messages.PROJECT_HELP_LABELS__NAME)
@click.option('-ln', '--label-name', prompt=messages.PROJECT_HELP_LABELS__LABEL_NAME,
              help=messages.PROJECT_HELP_LABELS__LABEL_NAME)
@prepare_command()
def add_label_command(cnvrg, logger, name, label_name):
    """
    Add label to project
    """
    project = cnvrg.projects.get(name)
    label = cnvrg.labels.get(label_name, LabelKind.PROJECT)
    project.labels.add(label)
    success_message = messages.PROJECT_ADD_LABEL_SUCCESS.format(label_name, name)
    logger.log_and_echo(success_message)


@project_group.command(name='remove-label')
@click.option('-n', '--name', prompt=messages.PROJECT_HELP_LABELS__NAME, help=messages.PROJECT_HELP_LABELS__NAME)
@prepare_command()
def remove_label_command(cnvrg, logger, name):
    """
    Remove label from a project
    """
    project = cnvrg.projects.get(name)
    project.labels.remove()
    success_message = messages.PROJECT_REMOVE_LABEL_SUCCESS.format(name)
    logger.log_and_echo(success_message)


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_PROMPT_DELETE, help=messages.PROJECT_HELP_DELETE)
@prepare_command()
def delete(cnvrg, logger, name):
    """
    Deletes the current project
    """
    click.confirm(messages.PROJECT_DELETE_CONFIRM.format(name), default=False, abort=True)
    cnvrg.projects.delete([name])
    success_message = messages.PROJECT_DELETE_SUCCESS.format(name)
    logger.log_and_echo(success_message)


@project_group.command()
@click.option('-g', '--git-diff', is_flag=True, help=messages.DATA_UPLOAD_HELP_GIT_DIFF)
@prepare_command()
def upload(project, logger, git_diff):
    """
    Uploads updated files from the current project folder
    """
    project.upload(progress_bar_enabled=True, git_diff=git_diff)
    logger.log_and_echo(messages.DATA_UPLOAD_SUCCESS)


@project_group.command()
@prepare_command()
def download(project, logger):
    """
    Downloads updated files to the current project folder
    """
    project.download(progress_bar_enabled=True)
    logger.log_and_echo(messages.DATA_DOWNLOAD_SUCCESS)


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_LINK_PROMPT_NAME, help=messages.PROJECT_LINK_HELP_NAME)
@prepare_command()
def link(cnvrg, logger, name):
    """
    Links the current directory with a new project.
    This command will create a new project and upload the current folder content to the newly created project.
    """
    curr_dir_project_name = Config().data_owner_slug
    if curr_dir_project_name:
        error_message = error_messages.DIRECTORY_ALREADY_LINKED.format(curr_dir_project_name)
        logger.log_and_echo(error_message, error=True)

    logger.log_and_echo(messages.PROJECT_CREATE_NEW.format(name))
    new_project = cnvrg.projects.create(name=name)
    new_project.local_commit = new_project.commit

    logger.log_and_echo(messages.PROJECT_CONFIGURING_FOLDER)
    new_project.save_config()

    logger.log_and_echo(messages.PROJECT_UPLOAD)
    new_project.upload(progress_bar_enabled=True)

    success_message = messages.PROJECT_LINK_SUCCESS.format(new_project.title)
    logger.log_and_echo(success_message)


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_LINK_GIT_PROMPT_NAME, help=messages.PROJECT_LINK_GIT_HELP_NAME)
@click.option('-s', '--soft', is_flag=True, help=messages.SOFT_LINK_GIT_PROJECT)
@prepare_command()
def link_git(cnvrg, logger, name, soft):
    """
     Links the current directory, which is expected to be a git directory, with an existing cnvrg project
     that is configured to be a git project.
    """
    git_proj = cnvrg.projects.get(name)
    if not git_proj:
        project_not_exist_error = error_messages.PROJECT_NOT_EXIST.format(name)
        logger.log_and_echo(project_not_exist_error, error=True)

    if not git_proj.git:
        error_message = error_messages.NOT_A_GIT_PROJECT.format(git_proj.title)
        logger.log_and_echo(error_message, error=True)

    if soft and os.path.exists(CONFIG_FOLDER_NAME):
        logger.log_and_echo(messages.PROJECT_GIT_LINK_SKIP.format(git_proj.title))
        return

    logger.log_and_echo(messages.PROJECT_CONFIGURING_FOLDER)
    git_proj.local_commit = git_proj.commit
    git_proj.save_config()

    if not cnvrgignore_exists(os.getcwd()):
        create_cnvrgignore(os.getcwd())

    output_folder = os.environ.get('CNVRG_OUTPUT_DIR')
    if output_folder and not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    success_message = messages.PROJECT_LINK_SUCCESS.format(git_proj.title)
    logger.log_and_echo(success_message)


@project_group.command()
@click.option('-j', '--job-slug', default=None, help=messages.PROJECT_JOB_SLUG_HELP_NAME)
@click.option('-g', '--git-diff', is_flag=True, default=None, help=messages.DATA_UPLOAD_HELP_GIT_DIFF)
@click.option('-m', '--message', default="", help=messages.DATA_COMMIT_MESSAGE)
@click.option('-o', '--output-dir', default=None, help=messages.SYNC_OUTPUT_DIR)
@click.option('-d', '--debug-mode', is_flag=True, default=False, help=messages.DEBUG_MODE_HELP)
@prepare_command()
def sync(project, logger, message, job_slug=None, git_diff=None, output_dir=None, debug_mode=None):
    """
    Sync local project to remote
    """
    logger.log_and_echo('Syncing Project')

    # store the last commit before syncing
    start_commit = project.last_commit
    commit_sha1 = project.sync(job_slug=job_slug, git_diff=git_diff, progress_bar_enabled=True, message=message,
                               output_dir=output_dir, debug_mode=debug_mode)
    # if the last commit is the same as the start commit before syncing, the project is up to date
    if commit_sha1 is None or commit_sha1 == start_commit:
        logger.log_and_echo(messages.PROJECT_IS_UP_TO_DATE.format(project.slug))
    else:
        logger.log_and_echo(messages.PROJECT_SYNC_SUCCESS.format(project.slug, commit_sha1))


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_PROMPT_NAME, help=messages.PROJECT_HELP_NAME)
@click.option('-f', '--files', prompt=messages.DATASET_PUT_PROMPT_FILES, help=messages.DATASET_PUT_HELP_FILES)
@click.option('-fc', '--force', is_flag=True, help=messages.DATA_UPLOAD_HELP_FORCE)
@click.option('-or', '--override', is_flag=True, help=messages.DATA_UPLOAD_HELP_OVERRIDE)
@click.option('-gd', '--git-diff', is_flag=True, help=messages.DATA_UPLOAD_HELP_GIT_DIFF)
@click.option('-d', '--dir', default="", help=messages.DATA_LOCATION_IN_STORAGE)
@prepare_command()
def put(cnvrg, logger, name, files, force, override, git_diff, dir):
    """
    Uploads the given files to the given project
    """
    file_paths = files.split(",")
    project = cnvrg.projects.get(name)
    sha1 = project.put_files(
        paths=file_paths,
        progress_bar_enabled=True,
        git_diff=git_diff,
        force=force,
        override=override,
        dir_path=dir
    )
    logger.log_and_echo("\n" + messages.FILES_UPLOAD_SUCCESS)
    logger.log_and_echo(messages.COMMIT_SHA1_MESSAGE.format(sha1))


@project_group.command()
@click.option('-n', '--name', prompt=messages.PROJECT_PROMPT_NAME, help=messages.PROJECT_HELP_DELETE)
@click.option('-f', '--files', prompt=messages.DATASET_REMOVE_PROMPT_FILES, help=messages.DATASET_REMOVE_HELP_FILES)
@click.option('-m', '--message', help=messages.DATA_COMMIT_MESSAGE, default="")
@prepare_command()
def rm(cnvrg, logger, name, files, message):
    """
    Removes the given files remotely
    """
    file_paths = files.split(",")
    project = cnvrg.projects.get(name)
    removed_files_count = project.remove_files(paths=file_paths, message=message, progress_bar_enabled=True)
    str_count = str(removed_files_count)
    pluralize = "File{}".format('s'[:removed_files_count ^ 1])  # Pluralizes the word (File/Files)
    logger.log_and_echo("{} {} removed.".format(str_count, pluralize))



import re
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.config import error_messages

NAME_VALIDATION_REGEX = re.compile(r"\A[a-zA-Z0-9\-_\/]+\Z")


class LabelKind:
    PROJECT = "projects"
    DATASET = "datasets"

    @classmethod
    def validate(cls, kind):
        return kind in {cls.PROJECT, cls.DATASET}


class LabelColor:
    BLUE = 'Blue'
    GREEN = 'Green'
    LIGHT_GREEN = 'Light green'
    LIGHT_BLUE = 'Light blue'
    YELLOW = 'Yellow'
    RED = 'Red'
    PURPLE = 'Purple'
    LIGHTGREY = 'Light grey'
    GREY = 'Grey'

    ALL_COLORS = {BLUE, GREEN, LIGHT_GREEN, LIGHT_BLUE, YELLOW, RED, PURPLE, LIGHTGREY, GREY}

    @classmethod
    def validate(cls, color_name):
        return color_name in cls.ALL_COLORS


def validate_labels_params(name, kind, color_name=None):
    if not name or not isinstance(name, str) or not re.match(NAME_VALIDATION_REGEX, name):
        raise CnvrgArgumentsError(error_messages.LABEL_BAD_NAME)

    if not LabelKind.validate(kind):
        raise CnvrgArgumentsError(error_messages.LABEL_BAD_KIND)

    if color_name and not LabelColor.validate(color_name):
        raise CnvrgArgumentsError(error_messages.LABEL_BAD_COLOR)



import ast
import click


class PythonLiteralOption(click.Option):

    def type_cast_value(self, ctx, value):
        try:
            if value is None:
                return None
            if isinstance(value, list) and not value:
                return []
            if isinstance(value, dict) and not value:
                return {}
            return ast.literal_eval(value)
        except Exception:
            raise click.BadParameter(value)



import time
from contextlib import contextmanager
from datetime import datetime

from cnvrgv2.modules.flows.flow import Flow
from cnvrgv2.utils.cron import Cron


class TestFlows:
    # TODO: Need to add cleanup by context + add create flow with title (in server)

    @staticmethod
    @contextmanager
    def temp_flow(project, **kwargs):
        flow = project.flows.create(**kwargs)
        yield flow
        flow.delete()

    def test_create_flow_success(self, e2e_project):
        with TestFlows.temp_flow(e2e_project) as flow:
            assert type(flow) is Flow

    def test_create_flow_with_yaml_success(self, e2e_project, e2e_flow_yaml):
        with TestFlows.temp_flow(e2e_project, yaml_path=e2e_flow_yaml["path"]) as flow:
            assert flow.title == e2e_flow_yaml["title"]

    def test_create_flow_with_yaml_string_success(self, e2e_project, e2e_flow_yaml):
        with TestFlows.temp_flow(e2e_project, yaml_string=e2e_flow_yaml["string"]) as flow:
            assert flow.title == e2e_flow_yaml["title"]

    def test_run_flow_success(self, e2e_project, e2e_flow_yaml):
        with TestFlows.temp_flow(e2e_project, yaml_path=e2e_flow_yaml["path"]) as flow:
            flow.run()

            flow_version = next(flow.flow_versions.list())
            experiments = e2e_project.experiments.list()

            assert flow_version.state == "running"
            assert len(list(experiments)) == 1

    def test_get_flow_success(self, e2e_project, e2e_flow):
        slug = e2e_flow.slug
        flow = e2e_project.flows.get(slug)
        assert slug == flow.slug

    def test_update_flow_success(self, random_name, e2e_flow):
        title = random_name(5)
        flow = e2e_flow.update(title=title)
        assert title == flow.title

    # DEV-10527: Currently disable. Missing full implementation in the server
    # def test_delete_flow_success(self, e2e_project, e2e_flow):
    #    slug = e2e_flow.slug
    #    e2e_flow.delete()
    #
    #    with pytest.raises(CnvrgHttpError) as exception_info:
    #        e2e_project.flows.get(slug).title
    #
    #    assert exception_info.value.status_code == 404

    def test_list_flows_only_success(self, e2e_project):
        num_of_flows = 5

        # Create some flows
        for i in range(num_of_flows):
            e2e_project.flows.create()

        flows_list = e2e_project.flows.list()

        assert len(list(flows_list)) == num_of_flows

        for flow in flows_list:
            flow.delete()

    def test_list_flows_sort_success(self, e2e_project):
        num_of_flows = 2
        date_format = "%Y-%m-%dT%H:%M:%S.%fZ"

        # Create some flows
        for i in range(num_of_flows):
            e2e_project.flows.create()
            time.sleep(1)

        flows = e2e_project.flows.list("-created_at")
        first_date = datetime.strptime(next(flows).created_at, date_format)
        second_date = datetime.strptime(next(flows).created_at, date_format)

        assert first_date > second_date

        for flow in flows:
            flow.delete()

    # DEV-10527: Currently disable. Missing full implementation in the server
    # def test_bulk_delete_flows_success(self, e2e_clean_flows, e2e_project):
    #    num_of_flows = 2
    #    flows = []
    #
    #    # Create some flows
    #    for i in range(num_of_flows):
    #        flow = TestFlows.create_flow(e2e_project)
    #        flows.append(flow)
    #
    #    slugs = list(flow.slug for flow in flows)
    #    e2e_project.flows.delete(slugs)
    #
    #    for slug in slugs:
    #        with pytest.raises(CnvrgHttpError) as exception_info:
    #            e2e_project.flows.get(slug).title
    #        assert exception_info.value.status_code == 404

    def test_flow_set_schedule_success(self, e2e_flow):
        cron = Cron(minute=30, hour=1)

        e2e_flow.set_schedule(str(cron))

        e2e_flow.reload()
        assert e2e_flow.cron_syntax == str(cron) + " UTC"

    def test_flow_clear_schedule_success(self, e2e_flow):
        cron = Cron(minute=30, hour=1)
        e2e_flow.set_schedule(str(cron))

        e2e_flow.clear_schedule()

        e2e_flow.reload()
        assert e2e_flow.cron_syntax is None

    def test_toggle_webhook_on_success(self, e2e_flow):
        e2e_flow.toggle_webhook(True)

        e2e_flow.reload()
        assert e2e_flow.webhook_url

    def test_toggle_webhook_off_success(self, e2e_flow):
        e2e_flow.toggle_webhook(True)
        e2e_flow.toggle_webhook(False)

        e2e_flow.reload()
        assert e2e_flow.webhook_url is None

    def test_toggle_dataset_update_on_success(self, e2e_flow, e2e_dataset):
        e2e_flow.toggle_dataset_update(True, e2e_dataset)

        e2e_flow.reload()
        assert e2e_flow.trigger_dataset

    def test_toggle_dataset_update_off_success(self, e2e_flow, e2e_dataset):
        e2e_flow.toggle_dataset_update(True, e2e_dataset)
        e2e_flow.toggle_dataset_update(False)

        e2e_flow.reload()
        assert e2e_flow.trigger_dataset is None



from tests.e2e.conftest import call_database
from tests.e2e.data_owner_e2e.data_owner_labels import DataOwnerLabels


class TestDatasetLabels(DataOwnerLabels):
    @staticmethod
    def cleanup(context_prefix):
        delete_command = "DELETE FROM labels WHERE name LIKE'{}%'".format(context_prefix)
        call_database(delete_command)

    def test_list_labels_of_dataset(self, e2e_dataset, e2e_label_dataset):
        self.data_owner_list_labels(e2e_dataset, e2e_label_dataset)

    def test_add_labels_to_dataset(self, e2e_dataset, e2e_label_dataset):
        self.data_owner_add_labels(e2e_dataset, e2e_label_dataset)

    def test_remove_labels_from_dataset(self, e2e_dataset, e2e_label_dataset):
        self.data_owner_remove_labels(e2e_dataset, e2e_label_dataset)



from cnvrgv2.modules.queries.query import Query


class TestQueries:
    # Tests for fullpath queries. To test tag queries we need to add support for elastic

    @staticmethod
    def create_query(dataset, name, query_raw):
        return dataset.queries.create(name=name, query=query_raw)

    def test_create(self, random_name, e2e_dataset):
        name = random_name(5)
        model = TestQueries.create_query(dataset=e2e_dataset, name=name, query_raw="fake-query")
        assert type(model) == Query
        assert model.name == name




class TestFlowVersion:
    # TODO: add more tests probably?

    def test_get_flow_version_success(self, e2e_flow, e2e_flow_version):
        title = e2e_flow_version.title
        fv = e2e_flow.flow_versions.get(title)
        assert title == fv.title

    def test_list_flow_versions_success(self, e2e_flow):
        fvs = e2e_flow.flow_versions.list()
        assert len(list(fvs)) > 0

    def test_stop_flow_version_success(self, e2e_flow):
        complete_state = "completed"

        e2e_flow.run()
        fv = next(e2e_flow.flow_versions.list())
        fv.stop()
        fv.reload()

        assert fv.state == complete_state

    def test_info_flow_version_success(self, e2e_flow_version):
        fv_status = 'fv_status'
        info = e2e_flow_version.info()
        assert fv_status in info.keys()



class Operators:
    LIKE = "LIKE"
    GT = ">"
    LT = "<"
    IS = "="


def list_to_multiple_conditions(key, operator, values):
    """
    Will create a list with multiple condition objects
    @param key: The key that will be presented in each condition
    @param operator: The operator to apply in the condition
    @param values: list of values to convert to conditions
    @return: List of conditions
    """
    conditions_list = []
    template_condition = {
            "key": key,
            "operator": operator
        }

    for value in values:
        current_condition = dict(template_condition)
        current_condition["value"] = value
        conditions_list.append(current_condition)

    return conditions_list



from cnvrgv2.config import routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.casters import safe_cast
from cnvrgv2.utils.validators import attributes_validator, validate_gpu


class KubeTemplate(DynamicAttributes):
    available_attributes = {
        "slug": str,
        "title": str,
        "cpu": float,
        "memory": float,
        "users": list,
        "gpu": dict,
        "mig_device": str,
        "spark_config": str,
        "gaudi": float,
        "is_private": bool,
        "compute_type": str,
        "labels": list,
        "taints": list,
        "worker": dict,
        "cluster_title": str,
        "num_executors": int,
        "templates_ids": list
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current template
        if slug:
            self._context.set_scope(SCOPE.TEMPLATE, slug)

        scope = self._context.get_scope(SCOPE.TEMPLATE)

        self._proxy = Proxy(context=self._context)
        self._route = routes.TEMPLATE_BASE.format(
            scope["organization"],
            'clusters',
            scope['resource'],
            scope["template"]
        )
        self._attributes = attributes or {}
        self.slug = scope["template"]

    def update(self, **kwargs):
        """
        Updates given attributes of a kube compute template

        @param kwargs: A list of optional attributes to update
            title: Name of the kube template
            cpu: CPU cores of the kube template
            memory: RAM memory of the kube template
            users: Users with access to the kube template
            gpu: GPU cores of the kube template
            shm: Shared memory (RAM) count of the kube template
            huge_pages: Gaudi huge pages count of the kube template
            gaudi: Gaudi accelerators num of the kube template
            is_private: is the kube template private or public
            spark_config: A dict of spark executor properties
            mig_device: mig device type for the kube template
            labels: labels of the kube template
            taints: taints of the kube template
            worker_num_executors: num of workers for the kube template
            worker_cpu: worker CPU cores of the kube template
            worker_memory: worker RAM memory of the kube template
            worker_huge_pages: worker Gaudi huge pages count of the kube template
            worker_gpu: worker GPU cores of the kube template
            worker_mig_device: worker mig device type
            worker_labels: worker labels
            worker_taints: worker taints

        """

        updated_attributes = {
            **kwargs,
        }

        parent_attributes = {
            "cpu": float,
            "memory": float,
            "labels": str,
            "taints": str,
            "gaudi": float,
            "shm": float,
            "huge_pages": float
        }

        for key in parent_attributes:
            if key in kwargs:
                updated_attributes['parent_' + key] = safe_cast(kwargs[key], parent_attributes[key], kwargs[key])
                updated_attributes.pop(key, None)

        if kwargs.get("gpu") is not None or kwargs.get("mig_device"):
            updated_attributes["parent_gpu"] = {
                "count": float(kwargs.get("gpu")) if kwargs.get("gpu") is not None
                else self._attributes["gpu"],
                "mig_device": kwargs.get("mig_device") if kwargs.get("mig_device") is not None
                else self._attributes["mig_device"]
            }
            updated_attributes.pop('mig_device', None)
            updated_attributes.pop('gpu', None)
            validate_gpu(updated_attributes["parent_gpu"])

        if kwargs.get("worker_gpu") or kwargs.get("worker_mig_device"):
            updated_attributes["worker_gpu"] = {
                "count": kwargs.get("worker_gpu") if kwargs.get("worker_gpu")
                else self._attributes["worker"]["gpu"],
                "mig_device": kwargs.get("worker_mig_device") if kwargs.get("worker_mig_device")
                else self._attributes["worker"]["mig_device"]
            }
            updated_attributes.pop('worker_gpu', None)
            updated_attributes.pop('worker_mig_device', None)
            validate_gpu(updated_attributes["worker_gpu"])

        if "is_private" in kwargs:
            updated_attributes["is_private"] = kwargs["is_private"]

        attributes_validator(
            {**self.available_attributes,
             "parent_cpu": float,
             "parent_memory": float,
             "parent_labels": str,
             "parent_taints": str,
             "parent_gpu": dict,
             "worker_gpu": dict,
             "worker_num_executors": int,
             "worker_cpu": float,
             "worker_huge_pages": float,
             "worker_memory": float,
             "worker_shm": float,
             "worker_labels": str,
             "worker_taints": str,
             "parent_shm": float,
             "parent_huge_pages": float,
             "parent_gaudi": float,
             }, updated_attributes,
        )

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.PUT,
            payload={"template_slug": self.slug, **updated_attributes})

        self._attributes = response.attributes

    def save(self):
        """
        In case of any attribute change, saves the changes
        """
        self.update(**self._attributes)

    def delete(self):
        """
        Deletes the current template
        @return: None
        """
        self._proxy.call_api(route=self._route, http_method=HTTP.DELETE)



import re

from cnvrgv2.config import error_messages
from cnvrgv2.errors import CnvrgArgumentsError

URL_VALIDATION_REGEX = re.compile(r"\A[\w.-]+(?:\.[\w\.-]+)+[\w\-\._%~:\/?#\[\]@!\$&'\(\)\*\+,;=.]+\Z")


class RegistryTypes:
    CNVRG = "cnvrg"
    INTEL = "intel"
    HABANA = "habana"
    DOCKERHUB = "dockerhub"
    NVIDIA = "nvidia"
    GCR = "gcr"
    ECR = "ecr"
    ACR = "acr"
    OTHER = "other"

    @classmethod
    def validate_type(cls, registry_type):
        if registry_type is None:
            return True

        registry_types = [cls.__dict__[var] for var in vars(cls) if not var.startswith("__")]
        return registry_type in registry_types


def validate_registry_params(title=None, url=None, username=None, password=None):
    if title is not None and not isinstance(title, str):
        raise CnvrgArgumentsError(error_messages.REGISTRY_BAD_TITLE)

    if url is not None and not isinstance(url, str):
        raise CnvrgArgumentsError(error_messages.REGISTRY_BAD_URL)

    if isinstance(url, str) and not re.match(URL_VALIDATION_REGEX, url):
        raise CnvrgArgumentsError(error_messages.REGISTRY_BAD_URL_FORMAT)

    if username is not None and not isinstance(username, str):
        raise CnvrgArgumentsError(error_messages.REGISTRY_BAD_USERNAME)

    if password is not None and not isinstance(password, str):
        raise CnvrgArgumentsError(error_messages.REGISTRY_BAD_PASSWORD)



from cnvrgv2.modules.volumes.volume import Volume
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.config import error_messages, routes
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.url_utils import urljoin


class VolumesClient:
    def __init__(self, project):
        self._context = Context(context=project._context)
        scope = self._context.get_scope(SCOPE.PROJECT)

        self._proxy = Proxy(context=self._context)
        self._route = routes.VOLUMES_BASE.format(scope["organization"], scope["project"])

    def list(self, sort="-id"):
        """
        List all volumes in a specific organization
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields volume objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=Volume,
            sort=sort
        )

    def get(self, slug=None, name=None):
        """
        Retrieves an volume by the given slug, or name
        @param slug: The slug of the requested volume
        @param name: The name of the requested volume
        @return: volume object
        """

        if slug and isinstance(slug, str):
            return Volume(context=self._context, slug=slug)
        elif not slug and name:
            get_volume_by_name_url = urljoin(self._route, routes.GET_BY_NAME_SUFFIX)

            attributes = {
                "volume_name": name,
            }

            res_attributes = self._proxy.call_api(
                route=get_volume_by_name_url,
                http_method=HTTP.GET,
                payload=attributes
            ).attributes

            return Volume(context=self._context, slug=res_attributes['slug'], attributes=res_attributes)

        else:
            raise CnvrgArgumentsError(error_messages.VOLUME_GET_FAULTY_PARAMS)

    def create(self, title, size, templates=None):
        """
        Creates a new volume
        @param title: Title of the volume
        @param size: Size of the volume
        @param templates:
        @return:
        """
        attributes = {
            "template_names": templates,
            "external_disk_title": title,
            "external_disk_size": size,
        }

        res_attributes = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="Volume", attributes=attributes)
        ).attributes

        return Volume(context=self._context, slug=res_attributes["slug"], attributes=res_attributes)



from cnvrgv2 import cnvrg
from cnvrgv2.config import error_messages, routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.labels.label import Label
from cnvrgv2.modules.labels.utils import LabelColor, LabelKind, validate_labels_params
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.json_api_format import JAF


class LabelsClient:
    def __init__(self, organization) -> None:
        if type(organization) is not cnvrg.Cnvrg:
            raise CnvrgArgumentsError(error_messages.ARGUMENT_BAD_TYPE.format(cnvrg.Cnvrg, type(organization)))

        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._scope = scope
        self._proxy = Proxy(context=self._context)
        self._route = routes.LABELS_BASE.format(scope["organization"])

    def list(self, kind, filter=None):
        """
        List all Labels in a specific organization
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields labels objects
        """
        if not LabelKind.validate(kind):
            raise CnvrgArgumentsError(error_messages.LABEL_BAD_KIND)

        def label_builder(item):
            attributes = item.attributes
            attributes['id'] = item.id
            return Label(self._context, name=item.attributes['name'], kind=item.attributes['kind'], attributes=item.attributes)

        res = api_list_generator(
            context=self._context,
            route=self._route,
            object_builder=label_builder,
            identifier='name',
            filter=filter,
            data={'kind': kind},
        )

        return res

    def create(self, name, kind, color_name=LabelColor.BLUE):
        """
        Create a new label
        @param name: [String] The label name
        @param kind: [String] The kind of the tag, either 'projects' or 'datasets' are available.
        @param color_name: [String] The name of the color from the available colors list
        @raise CnvrgHttpError: if label already exists.
        @return: [Label object] The newly created label
        """
        validate_labels_params(name, kind, color_name)

        attributes = {
            "name": name,
            "kind": kind,
            "color_name": color_name,
        }
        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="label", attributes=attributes)
        )

        name = response.attributes['name']
        kind = response.attributes['kind']
        return Label(self._context, name=name, kind=kind, attributes=response.attributes)

    def get(self, name, kind):
        """
        Retrieves a Label by the given name
        @param name: [String] The name of the label
        @return: Image object
        """
        validate_labels_params(name, kind)
        filter = {"operator": "AND", "conditions": [{"key": "name", "operator": "is", "value": name}]}
        generator = self.list(kind=kind, filter=filter)
        try:
            label = next(generator)
            return label
        except StopIteration:
            raise CnvrgArgumentsError(error_messages.LABEL_NOT_FOUND)



import os
import yaml
import shutil
import pytest

from contextlib import contextmanager

from cnvrgv2.config import Config
from cnvrgv2.errors import CnvrgError


class TestConfig:
    @staticmethod
    def destroy_config():
        config = Config()
        # Destroy global config
        if os.path.exists(config.global_cnvrg_path):
            shutil.rmtree(config.global_cnvrg_path)

        # Destroy local config
        if os.path.exists(config.local_cnvrg_path):
            shutil.rmtree(config.local_cnvrg_path)

    @staticmethod
    @contextmanager
    def create_config(path, content):
        """
        Creates a config file context manager with cleanup
        @param path: The path of the config file
        @param content: the content of the config file
        @return: None
        """
        with open(path, "w+") as f:
            yaml.dump(content, f)

        yield
        os.remove(path)

    def test_init_empty_config_files(self, mocker):
        mocker.patch.object(Config, "global_config_exists", False)
        mocker.patch.object(Config, "_load_legacy_config", lambda x: True)

        config = Config()
        for key in Config.global_config_attributes.keys():
            assert getattr(config, key) is None

        for key in Config.local_config_attributes.keys():
            assert getattr(config, key) is None

    def test_init_global_config_files(self, mocker):
        global_config = {
            "organization": "test-org",
            "domain": "test-domain",
            "token": "test-token",
            "user": "test-user",
            "check_certificate": "test-cert",
            "keep_duration_days": "test-dur",
            "version": "test-version",
            "sso_version": "v2"
        }

        # Set up that only global config is loaded
        mocker.patch.object(Config, "_load_legacy_config", lambda x: True)
        mocker.patch.object(Config, "_load_local_config", lambda x: True)

        # Create the global config
        global_config_path = os.getcwd() + "/.global.config"
        mocker.patch.object(Config, "global_config_exists", True)
        mocker.patch.object(Config, "global_config_file_path", global_config_path)

        with TestConfig.create_config(global_config_path, global_config):
            config = Config()

            for key in Config.global_config_attributes.keys():
                assert getattr(config, key) == global_config[key]

    def test_init_local_config_files(self, mocker):
        local_config = {
            "commit_sha1": "test-commit",
            "git": True,
            "project_slug": "test-project",
        }

        # Set up that only local config is loaded
        mocker.patch.object(Config, "_load_legacy_config", lambda x: True)
        mocker.patch.object(Config, "_load_global_config", lambda x: True)

        # Create the local config
        local_config_path = os.getcwd() + "/.local.config"
        mocker.patch.object(Config, "local_config_exists", True)
        mocker.patch.object(Config, "local_config_file_path", local_config_path)

        with TestConfig.create_config(local_config_path, local_config):
            config = Config()

            for key in local_config.keys():
                assert getattr(config, key) == local_config[key]

    def test_init_legacy_config_files(self, mocker, sample_netrc):
        # TODO: only checks local legacy config file
        legacy_config = {
            ":project_name": "test-commit",
            ":project_slug": "test-slug",
            ":owner": "test-org",
            ":git": True,
        }

        # Set up that only legacy config is loaded
        mocker.patch.object(Config, "_load_global_config", lambda x: True)
        mocker.patch.object(Config, "_load_local_config", lambda x: True)

        # Create the legacy config
        legacy_config_path = os.getcwd() + "/.legacy.config"
        mocker.patch.object(Config, "legacy_local_config_file_path", legacy_config_path)

        with TestConfig.create_config(legacy_config_path, legacy_config):
            config = Config()

            # Check fake netrc file
            assert config.user == "test@test.com"
            assert config.token == "sample_token"

            # Check config
            assert config.git == legacy_config[":git"]
            assert config.organization == legacy_config[":owner"]
            assert config.project_slug == legacy_config[":project_slug"]
            assert config.data_owner_slug == legacy_config[":project_slug"]
            assert config.dataset_slug is None

    def test_update_config_file(self, mocker):
        global_config = {
            "organization": "test-org",
            "domain": "test-domain",
        }
        local_config = {
            "commit_sha1": "test-commit",
            "git": "test-git",
        }

        test_global_cnvrg_path = os.path.join(os.path.expanduser("~"), ".cnvrg-test")
        mocker.patch("cnvrgv2.config.config.CONFIG_FOLDER_NAME", ".cnvrg-test")
        mocker.patch("cnvrgv2.config.config.GLOBAL_CNVRG_PATH", test_global_cnvrg_path)

        Config().update(**{**global_config, **local_config})

        try:
            config = Config()

            for key in global_config:
                assert getattr(config, key) == global_config[key]

            for key in local_config:
                assert getattr(config, key) == local_config[key]
        finally:
            # Clean files
            TestConfig.destroy_config()

    def test_save_config_file(self, mocker):
        global_config = {
            "organization": "test-org",
            "domain": "test-domain",
        }
        local_config = {
            "commit_sha1": "test-commit",
            "git": False,
        }

        mocker.patch("cnvrgv2.config.config.CONFIG_FOLDER_NAME", ".cnvrg-test")
        mocker.patch("cnvrgv2.config.config.GLOBAL_CNVRG_PATH", os.path.join(os.path.expanduser("~"), ".cnvrg-test"))

        try:
            config = Config()
            for key in global_config:
                setattr(config, key, global_config[key])
            for key in local_config:
                setattr(config, key, local_config[key])

            config.save()

            config = Config()
            for key in global_config:
                assert getattr(config, key) == global_config[key]

            for key in local_config:
                assert getattr(config, key) == local_config[key]
        finally:
            # Clean files
            TestConfig.destroy_config()

    def test_remove_config_fields(self, mocker):
        mocker.patch("cnvrgv2.config.config.CONFIG_FOLDER_NAME", ".cnvrg-test")
        mocker.patch("cnvrgv2.config.config.GLOBAL_CNVRG_PATH", os.path.join(os.path.expanduser("~"), ".cnvrg-test"))

        try:
            Config().update(**{"organization": "test-org"})

            config = Config()
            assert getattr(config, "organization") == "test-org"

            config.remove_config_fields("organization")

            config = Config()
            assert getattr(config, "organization") is None

        finally:
            # Clean files
            TestConfig.destroy_config()

    def test_remove_config_fields_invalid_field(self, mocker):
        config = Config()
        with pytest.raises(CnvrgError) as exception_info:
            config.remove_config_fields("fake")

        assert "fake" in str(exception_info.value)

    def test_get_credential_variables_empty(self, mocker):
        mocker.patch.object(Config, "_load_legacy_config", lambda x: True)
        mocker.patch("cnvrgv2.config.config.CONFIG_FOLDER_NAME", ".cnvrg-test")
        mocker.patch("cnvrgv2.config.config.GLOBAL_CNVRG_PATH", os.path.join(os.path.expanduser("~"), ".cnvrg-test"))

        config = Config()
        token, domain, user = config.get_credential_variables()
        assert token is False
        assert domain is False
        assert user is False

    def test_get_credential_variables_valid(self, mocker):
        global_config = {
            "domain": "test-domain",
            "token": "test-token",
            "user": "test-user",
        }

        mocker.patch.object(Config, "_load_legacy_config", lambda x: True)
        mocker.patch("cnvrgv2.config.config.CONFIG_FOLDER_NAME", ".cnvrg-test")
        mocker.patch("cnvrgv2.config.config.GLOBAL_CNVRG_PATH", os.path.join(os.path.expanduser("~"), ".cnvrg-test"))

        def mock_global_config_file(*args, **kwargs):
            return global_config

        mocker.patch.object(yaml, "safe_load", mock_global_config_file)
        mocker.patch.object(os.path, "exists", return_value=True)
        mocker.patch.object(Config, '_load_legacy_config', return_value=True)
        mocker.patch("builtins.open", return_value=True)

        config = Config()
        token, domain, user = config.get_credential_variables()
        assert token == "test-token"
        assert domain == "test-domain"
        assert user == "test-user"

    def test_get_data_owner_name(self, mocker):
        local_config = {"project_slug": "test-project"}

        def mock_local_config_file(*args, **kwargs):
            return local_config

        mocker.patch("cnvrgv2.config.config.CONFIG_FOLDER_NAME", ".cnvrg-test")
        mocker.patch("cnvrgv2.config.config.GLOBAL_CNVRG_PATH", os.path.join(os.path.expanduser("~"), ".cnvrg-test"))

        mocker.patch.object(yaml, "safe_load", mock_local_config_file)
        mocker.patch.object(os.path, "exists", return_value=True)
        mocker.patch.object(Config, '_load_legacy_config', return_value=True)
        mocker.patch("builtins.open", return_value=True)

        config = Config()
        assert config.project_slug == "test-project"
        assert config.data_owner_slug == "test-project"
